/*
 * cmd.c
 * functions for cancan's built-in #commands
 *
 * (created: Finn Arne Gangstad, Dec 25th, 1993)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>
#include <assert.h>
#ifdef AIX
#include <sys/select.h>
#endif
#include "cancan.h"
#include "cmd.h"
#include "term.h"
#include "edit.h"

static void show_aliases();
static void show_actions();
static void show_marks();
static void show_keybinds();
static void parse_bind();
static void norebind();
static void define_new_key();
static void print_seq();
static char *read_seq();
static int get_one_char();
static void delete_action();
static void change_action();
static void add_new_action();
static void show_attr_syntax();

static int cmd_help(), cmd_shell(), cmd_action(), cmd_alias(), cmd_bind();
static int cmd_cancel(), cmd_capture(), cmd_connect(), cmd_edit(), cmd_for();
static int cmd_hilite(), cmd_identify(), cmd_init(), cmd_map(), cmd_mark();
static int cmd_prefix(), cmd_quote(), cmd_read(), cmd_record(), cmd_retrace();
static int cmd_snoop(), cmd_speedwalk(), cmd_ver(), cmd_wrap(), cmd_zap();
#ifdef TELNETBUG
static int cmd_stdcolour();
#endif

extern int strncasecmp();

cmdstruct commands[] = 
{
   {"help", "[command|keys]\t\tthis text, or keyboard help", cmd_help},
   {"17", "command\t\t\trepeat 'command' 17 times", (int(*)())NULL},
#ifndef NOSHELL
   {"!", "shell-command\t\texecute a shell command using /bin/sh", cmd_shell},
#endif
   {"action", "[[+|-]name] [\"pattern\"=[action]]\tlist/delete/define actions",
	cmd_action},
   {"alias", "[name[=[text]]]\t\tlist/delete/define aliases", cmd_alias},
   {"bind", "[name[=[command]]]\tlist/delete/define key bindings", cmd_bind},
   {"cancel", "[number]\t\tcancel editing session", cmd_cancel},
   {"capture", "[filename]\t\tbegin/end of capture to file", cmd_capture},
   {"connect", "session-id [initstr]\tconnect a new session", cmd_connect},
   {"edit", "\t\t\t\tlist editing sessions", cmd_edit},
   {"for", "a b [c] cmd\t\titerate 'cmd' from a to b step c", cmd_for},
   {"hilite", "[attr]\t\t\thighlight your input line", cmd_hilite},
   {"identify", "[startact endact]\tsend MUME client identification",
	cmd_identify},
   {"init", "[-][[+]string]\t\tsend string to host on connect", cmd_init},
   {"map", "[-[number]]\t\tshow/clear (auto)map", cmd_map},
   {"mark", "[string[=[attr]]]\t\tlist/define markers", cmd_mark},
   {"prefix", "string\t\t\tprefix all lines with string", cmd_prefix},
   {"quote", "\t\t\t\ttoggle verbatim-flag", cmd_quote},
   {"read", "file\t\t\tread lines from file/pipe and send", cmd_read},
   {"record", "[name]\t\t\trecord typed lines into alias", cmd_record},
   {"retrace", "[number]\t\tretrace the last number steps", cmd_retrace},
   {"snoop", "session-id\t\ttoggle output display for session", cmd_snoop},
   {"speedwalk", "\t\t\ttoggle speedwalk mode", cmd_speedwalk},
#ifdef TELNETBUG
   {"stdcolour", "attr\t\t\tset standard background colour", cmd_stdcolour},
#endif
   {"ver", "\t\t\t\tshow cancan version", cmd_ver},
   {"wrap", "\t\t\t\tturn word wrapping on/off", cmd_wrap},
   {"zap", "session-id\t\t\tdisconnect a session", cmd_zap},
   {(char*)NULL, (char*)NULL, (int(*)())NULL}
};

/* anyone knows if ANSI 6429 talks about more than 8 colours? */
static char *colournames[] = {
    "black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"
};

char *recording_name;		/* name of recording alias or NULL */

static int cmd_record(arg)
char *arg;
{
    if(*arg) {
	if(recording_name) {
	    printf("#already recording to alias '%s'\n", recording_name);
	    free(recording_name);
	}
	strtok(arg, " ");	/* strip any accidental trailing blanks */
	recording_name = strdup(arg);
	printf("#recording into alias '%s'\n", recording_name);
    } else {
	if(recording_name) {
	    printf("#recording of alias '%s' stopped.\n", recording_name);
	    free(recording_name);
	    recording_name = NULL;
	} else {
	    printf("#not recording any alias\n");
	}
    }
    return 1;
}

static int cmd_init(arg)
char *arg;
{
  if (*arg == '-') 
   {
     initstr[0] = 0;
     printf("#init cleared.\n");
   }
  else 
   {
     if (*arg == '+')
     	strcat(initstr, arg+1);
     else if (*arg)
     	strcpy(initstr, arg);
     printf("#init: %s\n", initstr);
   }
  save_settings();
  return 1;
}

static int cmd_hilite(arg)
char *arg;
{
    int attr = parse_attributes(arg);
    if(attr < 0) {
	printf("#attribute syntax error\n");
	show_attr_syntax();
    } else {
	attr_string(attr, edattrbeg, edattrend);
	printf("#input highlighting is now %so%s%s.\n", 
	       edattrbeg, (attr == NOATTRCODE) ? "ff" : "n", edattrend);
    }
    return 1;
}

/*
 * return token number for attribute word:
 * this is colour number for colours, TOK_xxx for effects, TOK_ON for "on",
 * or -1 for unrecognized word.
 */
int attr_token(word)
char *word;
{
    int strcasecmp();
    int i;
    for(i = 0; i < COLOURS && strcasecmp(word, colournames[i]) != 0; i++);
    if(i < COLOURS) return i;
    if(strcasecmp(word, "bold") == 0)
	return BOLD;
    else if(strcasecmp(word, "underline") == 0)
	return UNDERLINE;
    else if(strcasecmp(word, "inverse") == 0
	    || strcasecmp(word, "reverse") == 0)
	return INVERSE;
    else if(strcasecmp(word, "on") == 0)
	return TOK_ON;
    else
	return TOK_INVALID;
}

/*
 * parse attribute description in line.
 * Return attribute if successful, -1 otherwise.
 */
int parse_attributes(line)
char *line;
{
    char *p;
    int t, fg;
    p = strtok(line, " ");
    
    if(p == NULL) {
	return NOATTRCODE;
    }
    if((t = attr_token(p))== TOK_INVALID) return -1;
    if(t < COLOURS) {
	fg = t;
	if((p = strtok(NULL, " ")) == NULL) return ATTRCODE(fg, NOATTRIB);
	if((t = attr_token(p)) != TOK_ON) return -1;
    } else if(t != TOK_ON) {
	return ATTRCODE(t, NOATTRIB);
    } else
	fg = NOATTRIB;
    if((p = strtok(NULL, " ")) == NULL) return -1;
    t = attr_token(p);
    if(t < 0 || t >= COLOURS) return -1;
    return ATTRCODE(fg, t);
}

/*
 * display attribute syntax
 */
static void show_attr_syntax()
{
    int i;
    printf("#attribute syntax:\n\tOne of:\tbold, underline, inverse\n");
    printf("\tor:\t[foreground] [ON background]\n");
    printf("\tColours: ");
    for(i = 0; i < COLOURS; i++)
	printf("%s%c", colournames[i], (i == COLOURS - 1) ? '\n' : ',');
}

/*
 * return a static pointer to name of given attribute code
 */
char *attr_name(attrcode)
int attrcode;
{
    static char name[80];	/* should be enough */
    int fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);

    name[0] = 0;
    if(fg < COLOURS)
	strcpy(name, colournames[fg]);
    else if(fg == BOLD)
	strcpy(name, "bold");
    else if(fg == UNDERLINE)
	strcpy(name, "underline");
    else if(fg == INVERSE)
	strcpy(name, "inverse");
    if(bg != NOATTRIB) {
	if(name[0]) strcat(name, " ");
	strcat(name, "on ");
	assert(bg < COLOURS);
	strcat(name, colournames[bg]);
    }
    return name;
}

/*
 * put escape sequences to turn on/off an attribute in given buffers
 */
void attr_string(attrcode, begin, end)
int attrcode; char *begin, *end;
{	
    int fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);
    *begin = *end = '\0';
    if(fg < COLOURS) {
	if(bg < COLOURS)
	    sprintf(begin, "\033[3%d;4%dm", fg, bg);
	else
	    sprintf(begin, "\033[3%dm", fg);
    } else if(bg < COLOURS) {
	sprintf(begin, "\033[4%dm", bg);
    } else if(fg == BOLD) {
	if(modebold[0])
	    strcpy(begin, modebold);
	else {
	    strcpy(begin, modestandon);
	    strcpy(end, modestandoff);
	}
    } else if(fg == UNDERLINE) {
	if(modeuline[0])
	    strcpy(begin, modeuline);
	else {
	    strcpy(begin, modestandon);
	    strcpy(end, modestandoff);
	}
    } else if(fg == INVERSE) {
	if(modeinv[0])
	    strcpy(begin, modeinv);
	else {
	    strcpy(begin, modestandon);
	    strcpy(end, modestandoff);
	}
    }
    if(attrcode != NOATTRCODE && !*end)
	strcpy(end, modenorm);
}

static int cmd_mark(arg)
char *arg;
{
    if(!*arg)
	show_marks();
    else
	parse_marker(arg, 0);
    return 1;
}

static int cmd_bind(arg)
char *arg;
{
    if(!*arg)
	show_keybinds();
    else
	parse_bind(arg);
    return 1;
}

/*
 * list keyboard bindings
 */
static void show_keybinds()
{
    keynode *p;
    int count = 0;
    for(p = keydefs; p; p = p->next) {
	if(p->funct == key_run_command) {
	    if(count == 0)
		printf("#user-defined keys:\n");
	    printf("#bind %s=%s\n", p->name, p->call_data);
	    count++;
	}
    }
    if(count == 0)
	printf("#No key bindings defined right now.\n");
}

/*
 * parse the #bind command from a definition file
 */
void parse_bind_from_file(arg)
char *arg;
{
    char keyseq[CAPLEN];
    char *q = keyseq;
    char *p;
    char *name = arg;
    char *command;
    int escape = 0;
    int caret = 0;
    p = strchr(arg, ' ');
    if(!p) {
	printf("#bad definition file syntax: #bind %s\n", arg);
	return;
    }
    *p++ = '\0';
    p = skipspace(p);
    for(; *p; p++) {
	if(escape) {
	    *q++ = *p;
	    escape = 0;
	} else if(caret) {
	    *q++ = CONTROL(*p);
	    caret = 0;
	} else if(*p == '\\')
	    escape = 1;
	else if(*p == '^')
	    caret = 1;
	else if(*p == '=')
	    break;
	else
	    *q++ = *p;
    }
    if(!*p) {
	printf("#bad definition file syntax\n");
	return;
    }
    *q = '\0';
    command = p + 1;
    add_keynode(name, keyseq, key_run_command, command);
}

/*
 * parse the argument of the #bind command (interactive)
 */
static void parse_bind(arg)
char *arg;
{
    char *p, *command;
    char *name = arg;
    keynode **np;
    p = strchr(arg, '=');
    if(p) {
	*p++ = '\0';
	np = lookup_key(name);
	if(*p) {
	    command = p;
	    if(*np) {
		if((*np)->funct == key_run_command) {
		    free((*np)->call_data);
		    (*np)->call_data = strdup(command);
		    printf("#redefined key: %s=%s\n", name, command);
		} else
		    norebind(name);
	    } else
		define_new_key(name, command);
	} else {
	    if(*np) {
		if((*np)->funct == key_run_command) {
		    printf("#key binding deleted: %s=%s\n", name,
			   (*np)->call_data);
		    delete_keynode(np);
		} else
		    norebind(name);
	    } else
		printf("#no such key: '%s'\n", name);
	}
    } else {
	np = lookup_key(name);
	if(*np) {
	    if((*np)->funct == key_run_command) {
		char buf[LINELEN];
		sprintf(buf, "#bind %s=%s", name,
			escape_specials((*np)->call_data));
		to_input_line(buf);
	    } else
		norebind(name);
	} else
	    printf("#no such key: '%s'\n", name);
    }
}

static void norebind(name)
char *name;
{
    printf("#key '%s' is predefined and cannot be rebound.\n", name);
}

/*
 * define a new key, prompting user to press it
 */
static void define_new_key(name, command)
char *name, *command;
{
    char *keyseq;
    keynode *p;
    int seqlen;
    printf("#please press the key you want to redefine: ");
    fflush(stdout);
    print_seq(keyseq = read_seq()); putchar('\n');
    seqlen = strlen(keyseq);
    if(keyseq[0] >= ' ')
	printf("#That is not a redefinable key.\n");
    else {
	for(p = keydefs; p; p = p->next)
	    if(strncmp(p->sequence, keyseq, seqlen) == 0) {
		if(p->funct == key_run_command)
		    printf("Key already bound as: %s=%s\n",
			   p->name, p->call_data);
		else
		    printf("Key '%s' already bound as an editing key.\n",
			   p->name);
		return;
	    }
	add_keynode(name, keyseq, key_run_command, command);
	printf("#new key binding: %s=%s\n", name, command);
    }
}

/*
 * read a single escape sequence from the keyboard; return static pointer
 */
static char *read_seq()
{
    static char seq[CAPLEN];
    int i = 0;
    seq[i++] = get_one_char(0);
    while(i < CAPLEN - 1 && (seq[i] = get_one_char(KBD_TIMEOUT)) >= 0)
	i++;
    seq[i] = '\0';
    return seq;
}

/*
 * read a single character from stdin, with timeout in microseconds.
 * timeout == 0 means wait indefinitely (no timeout).
 * return char or -1 if timeout was reached.
 */
static int get_one_char(timeout)
int timeout;
{
    void bzero();
    int select();
    extern int errno;
    int err;
    fd_set fds;
    unsigned char c;
    struct timeval timeoutbuf;

    FD_ZERO(&fds);
    FD_SET(0, &fds);
    timeoutbuf.tv_sec = 0;
    timeoutbuf.tv_usec = timeout;
    while((err = select(2, &fds, NULL, NULL, timeout ? &timeoutbuf : NULL)) < 0
	  && errno == EINTR);
    if(err < 0)
	syserr("select");
    else if(err == 0)
	return -1;
    if(read(0, &c, 1) != 1) syserr("read");
    return c;
}

/*
 * print an escape sequence human-readably
 */
static void print_seq(seq)
char *seq;
{
    while(*seq) {
	if(*seq == '\033')
	    printf("esc ");
	else if(*seq == ' ')
	    printf("space ");
	else if(*seq < ' ')
	    printf("^%c ", *seq | '@');
	else
	    printf("%c ", *seq);
	seq++;
    }
}

/*
 * return a static pointer to escape sequence made printable (for use in
 * definition-files
 */
char *seq_name(seq)
char *seq;
{
    static char buf[LINELEN];
    char *p = buf;
    /*
     * rules: escape =\^ with \. control chars are written as ^X.
     */
    for(;*seq; seq++) {
	if(*seq == '=' || *seq == '\\' || *seq == '^')
	    *p++ = '\\';
	if(*seq < ' ') {
	    *p++ = '^';
	    *p++ = *seq | '@';
	} else
	    *p++ = *seq;
    }
    *p = '\0';
    return buf;
}

/*ARGSUSED*/
static int cmd_speedwalk(arg)
char *arg;
{
    printf("#speedwalk is now o%s.\n", (speedwalk ^= 1) ? "n" : "ff");
    return 1;
}

static int cmd_map(arg)
char *arg;
{
    if (!*arg)  /* show map */
	show_automap();
    else  /* remove some steps without walking */
	remove_steps(atoi(arg));
    return 1;
}

static int cmd_retrace(arg)
char *arg;
{
    retrace_steps(atoi(arg));
    return 1;
}

static int cmd_alias(arg)
char *arg;
{
    if(!*arg)
	show_aliases();
    else
	parse_alias(arg, 0);
    return 1;
}

static int cmd_action(arg)
char *arg;
{
    if(!*arg)
	show_actions();
    else
	parse_action(arg, 0);
    return 1;
}

static int cmd_capture(arg)
char *arg;
{
    if(!*arg) {
	if(capfile) {
	    fclose(capfile);
	    capfile = NULL;
	    printf("#end of capture to file.\n");
	} else {
	    printf("#capture to what file?\n");
	}
    } else {
	if(capfile) {
	    printf("#capture already active.\n");
	} else {
	    if((capfile = fopen(arg, "w")) == NULL) {
		printf("#error writing file %s\n", arg);
	    } else {
		printf("#capture to %s active, '#capture' ends.\n", arg);
	    }
	}
    }
    return 1;
}

static int cmd_prefix(arg)
char *arg;
{
    strcpy(prefixstr, arg);
    printf("#prefix %s.\n", *arg ? "set" : "cleared");
    return 1;
}

#ifndef NOSHELL
static int cmd_shell(arg)
char *arg;
{
    int system();
    if(!*arg)
	printf("#that's easy\n");
    else {
	reset_terminal();
	system(arg);
	set_terminal();
    }
    line0 = lines - 1;
    gotoxy(0, line0); printf("\n");
    return 1;
}
#endif

/*ARGSUSED*/
static int cmd_quote(arg)
char *arg;
{
    printf("#%s mode.\n", (verbatim ^= 1) ? "verbatim" : "normal");
    return 1;
}

static int cmd_read(arg)
char *arg;
{
    FILE *fp;
    char buf[LINELEN];
    if(*arg) {
	fp = (arg[0] == '|') ? popen(arg + 1, "r") : fopen(arg, "r");
	if(!fp) {
	    printf("#error opening %s\n", arg);
	    return 1;
	}
	while(fgets(buf, LINELEN, fp) != NULL)
	    send_to_host(buf, strlen(buf));
	if(arg[0] == '|') pclose(fp); else fclose(fp);
	return 0;
    } else {
	printf("#syntax: #read filename or #read |command\n");
	return 1;
    }
}

/*ARGSUSED*/
static int cmd_identify(arg)
char *arg;
{
    edit_start[0] = edit_end[0] = '\0';
    if (*arg != '\0') {
	char *p = strchr(arg, ' ');
	if(p) {
	    *p++ = '\0';
	    strcpy(edit_end, p);
	}
	strcpy(edit_start, arg);
    }
    send_to_host("~$#EI\n", 6);
    return 1;
}

/*ARGSUSED*/
static int cmd_edit(arg)
char *arg;
{
    editsess *sp;
    if(edit_sess) {
	for(sp = edit_sess; sp; sp = sp->next)
	    printf("# %s (%d)\n", sp->descr, sp->key);
    } else
	printf("#no active editors.\n");
    return 1;
}

static int cmd_cancel(arg)
char *arg;
{
    editsess *sp;
    if(!edit_sess) {
	printf("#no active sessions to cancel.\n");
    } else {
	if(*arg) {
	    for(sp = edit_sess; sp; sp = sp->next)
		if(atoi(arg) == sp->key) {
		    cancel_edit(sp);
		    break;
		}
	    if(!sp) printf("#no such session - use the session number\n");
	} else {
	    if(edit_sess->next)
		printf("#several sessions active, use #cancel <number>\n");
	    else {
		cancel_edit(edit_sess);
	    }
	}
    }
    return 1;
}

static int cmd_help(arg)
char *arg;
{
    int i;
    if(*arg == '#')
	arg++;
    if(*arg) {
	FILE *f = fopen(HELPFILE, "r");
	char line[LINELEN];
	char *p;
	if(f == NULL) {
	    printf("#Can't open help file %s\n", HELPFILE);
	} else {
	    while((p = fgets(line, LINELEN, f))
		  && (line[0] != '@'
		      || strncasecmp(line + 1, arg, strlen(arg)) != 0));
	    if(p) {
		while(fgets(line, LINELEN, f) != NULL
		      && line[0] != '@') {
		    printf("%s", line);
		}
	    } else {
		printf("#No entry for '%s' in the help file.\n", arg);
	    }
	}
    } else {
	printf("#commands availible:\n");
	for(i = 0; commands[i].name; i++)
	    printf("#%s %s\n", commands[i].name, commands[i].help);
	printf("Use '#help keys' for keyboard help.\n");
    }
    return 1;
}

static int cmd_for(arg)
char *arg;
{
    int from, to, step, n, a, b, i;
    char buf[LINELEN], istr[10][PARAMLEN];

    n = sscanf(arg, "%d %d %n%d %n", &from, &to, &a, &step, &b);
    switch(n) {
    case 2:
	step = 1;
	arg += a;
	break;
    case 3:
	arg += b;
	break;
    default:
	printf("#bad for syntax\n");
	return 1;
    }
    if(step > 0)
	for(i = from; i <= to; i += step) {
	sprintf(istr[0], "%d", i);
	subst_param(arg, buf, istr);
	printf("[%d:]", i); fflush(stdout);
	parse_input(buf, 1);
    } else if(step < 0)
	for(i = from; i >= to; i += step) {
	    sprintf(istr[0], "%d", i);
	    subst_param(arg, buf, istr);
	    printf("[%d:]", i); fflush(stdout);
	    parse_input(buf, 1);
	} else {
	    printf("#infinite loop\n");
	    return 1;
	}
    return 0;
}

#ifdef TELNETBUG
static int cmd_stdcolour(arg)
char *arg;
{
    int attrcode = parse_attributes(arg);
    if(attrcode < 0) {
	printf("#invalid attribute syntax\n");
	show_attr_syntax();
    } else {
	int bg = BACKGROUND(attrcode), fg = FOREGROUND(attrcode);
	if(fg >=COLOURS || bg >= COLOURS) {
	    printf("#please specify foreground and background colours\n");
	} else {
	    sprintf(modenorm, "\033[;3%d;4%dm", fg, bg);
	    printf("%s#standard colour set.\n", modenorm);
	    if(edattrend[0]) strcpy(edattrend, modenorm);
	}
    }
    return 1;
}
#endif

/*ARGSUSED*/
static int cmd_wrap(arg)
char *arg;
{
    printf("#word wrap %sabled.\n", (wrap ^= 1) ? "en" : "dis");
    return 1;
}

/*ARGSUSED*/
static int cmd_ver(arg)
char *arg;
{
    printver();
    return 1;
}

static int cmd_connect (arg)
char *arg;
{
#ifdef TERM
  printf("#connect: multiple sessions not supported in term version.\n");
#else  
  
  char *n, *s;
  if (!*arg)
     printf("#connect: no session name\n");
  else 
   {
     n = strtok(arg, " ");
     s = strtok(0, "\0");
     connect_sess(n, s);
   }
#endif /* term */  
  return 1;
}

static int cmd_zap (arg)
char *arg;
{
  if (!*arg)
     printf("#zap: no session name\n");
  else
     disconnect_sess(arg);
  return 1;
}

static int cmd_snoop (arg)
char *arg;
{
  if (!*arg)
     printf("#snoop: no session name\n");
  else
     snoop_sess(arg);
  return 1;
}

/*
 * parse the #alias command
 * (if fromfile is true, call came from definition file)
 */
void parse_alias(str, fromfile)
char *str; int fromfile;
{
    char *left, *right;
    aliasnode **np;
    char buf[LINELEN];
    char *p, *q;
  
    left = str = skipspace(str);
  
    while(*str!='=' && *str!=' ' && *str) str++;
  
    if(*str == '=') {
	*str = '\0';
	right = ++str;
	np = lookup_alias(left);
	if(!*skipspace(str)) {
	    /* delete alias */
	    if(*np) {
		printf("#deleting alias: %s=%s\n", left, (*np)->subst);
		delete_aliasnode(np);
	    } else {
		printf("#unknown alias: %s\n", left);
	    }
	} else {
	    /* add/redefine alias */
	
	    /* check for recursion (only direct recursion is checked) */
	    strcpy(buf, right);
	    for(p = strtok(buf, ";"); p; p = strtok((char*)NULL, ";")) {
		for(q = p; *q && *q != ' ' && *q != CMDSEP; q++);
		*q = '\0';
		if(strcmp(p, left) == 0) {
		    printf("#alias can't be defined by itself\n");
		    return;
		}
	    }
	    if(*np) {
		free((*np)->subst);
		(*np)->subst = strdup(right);
		if(!(*np)->subst) syserr("strdup");
	    } else {
		add_aliasnode(left, right);
	    }
	    if(!fromfile)
		printf("#new alias: %s=%s\n", left, right);
	}
	if(!fromfile)
	    save_settings();
    } else {
	/* show alias */
     
	*str = '\0';
	np = lookup_alias(left);
	if(*np) {
	    char buf[LINELEN];
	    sprintf(buf, "#alias %s=%s", left, escape_specials((*np)->subst));
	    to_input_line(buf);
	} else
	    printf("#unknown alias: %s\n", left);
    }
}

/*
 * parse the #action command
 * this function is too damn complex because of the hairy syntax. it should be
 * split up or rewritten as an fsm instead.
 * (if fromfile is true, call came from definition file)
 */
void parse_action(str, fromfile)
char *str; int fromfile;
{
    char *p, *label, *pattern, *commands;
    actionnode **np = NULL;
    char sign = 0;
  
    str = skipspace(str);
    if(*str == '+' || *str == '-')
	sign = *str++;
    if(*str != '"') {
	label = str;
	p = label + strcspn(label, " =");
	if(*p == '=') {
	    *p++ = '\0';
	    np = lookup_action(label);
	    if(!*np)
		printf("#no action labeled '%s'\n", label);
	    else {
		if(*p) {
		    change_action(*np, NULL, p);
		    if(sign) (*np)->active = (sign == '+');
		} else {
		    delete_action(np);
		}
		if(!fromfile) save_settings();
	    }
	    return;
	} else if(*p) {
	    *p = '\0';
	    np = lookup_action(label);
	    str = strchr(p + 1, '"');
	} else {
	    np = lookup_action(label);
	    if(*np) {
		if(sign) {
		    (*np)->active = (sign == '+');
		    printf("#action %s \"%s\" is now o%s\n", label,
			   (*np)->pattern, (sign == '+') ? "n" : "ff");
		    if(!fromfile) save_settings();
		} else {
		    char buf[LINELEN];
		    sprintf(buf, "#action %s \"%s\"=%s", label, (*np)->pattern,
			    escape_specials((*np)->commands));
		    to_input_line(buf);
		}
	    } else
		printf("#no action labeled '%s'\n", label);
	    return;
	}
    }
    if(str && *str == '"') {
	pattern = str + 1;
	p = strchr(pattern, '"');
	if(!p) {
	    printf("#invalid action syntax: missing \"\n");
	    return;
	}
	*p++ = '\0';
	if(*p != '=') {
	    printf("#invalid action syntax: missing =\n");
	    return;
	}
	commands = p + 1;
	if(*commands) {
	    if(np) {
		if(*np) {
		    change_action(*np, pattern, commands);
		    if(sign) (*np)->active = (sign == '+');
		} else
		    add_new_action(label, pattern, commands, sign != '-',
				   fromfile);
	    } else {
		np = lookup_action_pattern(pattern);
		if(*np) {
		    change_action(*np, NULL, commands);
		} else {
		    add_anonymous_action(pattern, commands, fromfile);
		}
	    }
	} else {
	    if(!np || !*np) {
		np = lookup_action_pattern(pattern);
		if(!*np) {
		    printf("#no action with pattern \"%s\"\n", pattern);
		    return;
		}
	    }
	    delete_action(np);
	}
	if(!fromfile) save_settings();
    } else
	printf("#invalid action syntax: missing \"\n");
}

/*
 * delete an action node
 */
static void delete_action(nodep)
actionnode **nodep;
{
    printf("#action %s \"%s\" deleted\n", (*nodep)->label, (*nodep)->pattern);
    delete_actionnode(nodep);
}

/*
 * change fields of an existing action node
 * pattern or commands can be NULL if no change
 */
static void change_action(node, pattern, commands)
actionnode *node; char *pattern, *commands;
{
    if(pattern) {
	free(node->pattern);
	node->pattern = strdup(pattern);
    }
    if(commands) {
	free(node->commands);
	node->commands = strdup(commands);
    }
    printf("#changed action: %s \"%s\"=%s\n", node->label, node->pattern,
	   node->commands);
}		

/*
 * add an action with numbered label
 */
void add_anonymous_action(pattern, commands, fromfile)
char *pattern, *commands; int fromfile;
{
    static last = 0;
    char label[16];
    do {
	sprintf(label, "%d", ++last);
    } while(*lookup_action(label));
    add_new_action(label, pattern, commands, 1, fromfile);
}

/*
 * create new action
 */
static void add_new_action(label, pattern, commands, active, fromfile)
char *label, *pattern, *commands; int active, fromfile;
{
    add_actionnode(pattern, commands, label, active);
    if(!fromfile)
	printf("#new action: %s \"%s\"=%s\n", label, pattern, commands);
}

/*
 * parse arguments to the #mark command
 * (if fromfile is true, call came from definition file)
 */
void parse_marker(str, fromfile)
char *str; int fromfile;
{
    char *p;
  
    str = skipspace(str);
    p = strchr(str, '=');
    if(p == NULL) {
	marknode **np = lookup_marker(str);
	if(*np && !fromfile) {
	    char buf[LINELEN];
	    sprintf(buf, "#mark %s=%s",
		    (*np)->pattern, attr_name((*np)->attrcode));
	    to_input_line(buf);
	} else {
	    printf("#unknown marker: '%s'\n", str);
	}
    } else {
	int attrcode;
	marknode **np;

	*p++ = '\0';
	np = lookup_marker(str);
	attrcode = parse_attributes(p);
	if(attrcode < 0) {
	    printf("#invalid attribute syntax\n");
	    show_attr_syntax();
	} else if(attrcode == NOATTRCODE) {
	    if(*np) {
		printf("#mark deleted: %s=%s\n",
		       (*np)->pattern, attr_name((*np)->attrcode));
		delete_marknode(np);
	    } else
		printf("#no mark '%s'\n", str);
	} else {
	    if(*np) {
		(*np)->attrcode = attrcode;
		if(!fromfile) printf("#changed");
	    } else {
		add_marknode(str, attrcode);
		if(!fromfile) printf("#new");
	    }
	    if(!fromfile)
		printf(" mark: %s=%s\n", str, attr_name(attrcode));
	}
	if(!fromfile) save_settings();
    }
}

/*
 * show defined marks
 */
static void show_marks()
{
    marknode *p;
    printf("#%s marker%s defined%c\n", markers ? "The following" : "No",
	   (markers && !markers->next) ? " is" : "s are",
	   markers ? ':' : '.');
    for(p = markers; p; p = p->next)
	printf("#mark %s=%s\n", p->pattern, attr_name(p->attrcode));
}

/*
 * show defined aliases
 */
static void show_aliases()
{
    aliasnode *p;
  
    printf("#%s alias%s defined%c\n", aliases ? "The following" : "No",
	   (aliases && !aliases->next) ? " is" : "es are",
	   aliases ? ':' : '.');
    for(p = aliases; p; p = p->next)
	printf("#alias %s=%s\n", p->name, p->subst);
}

/*
 * show defined actions
 */
static void show_actions()
{
    actionnode *p;
  
    printf("#%s action%s defined%c\n", actions ? "The following" : "No",
	   (actions && !actions->next) ? " is" : "s are", actions ? ':' : '.');
    for(p = actions; p; p = p->next)
	printf("#action %c%s \"%s\"=%s\n", p->active ? '+' : '-', p->label,
	       p->pattern, p->commands);
}

/*
 * add escapes to protect special characters from being escaped.
 * Return pointer to static string.
 */
char *escape_specials(str)
char *str;
{
    static char buf[LINELEN];
    char *p = buf;
    while(*str) {
	if(strchr(";&\\", *str) != NULL)
	    *p++ = '\\';
	*p++ = *str++;
    }
    *p = '\0';
    return buf;
}
