/*
 * tel.c -- telnet protocol communication module for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <errno.h>
#include <arpa/telnet.h>
#define TELOPT_NAWS 31
#include <arpa/inet.h>
#ifndef NEXT
#include <unistd.h>
#endif

#ifdef TERM
#include "client.h"
#endif /* TERM */ 

#include "cancan.h"
#include "term.h"

#ifdef TELOPTS
#define TELOPTSTR(n) ((n) > NTELOPTS ? "unknown" : telopts[n])
#endif

void bcopy(), bzero();


#ifdef NOBCOPY
#define bcopy(a,b,c) memcpy ((b), (a), (c))
#endif

#ifdef NOBZERO
#define bzero(a,b) memset ((a), 0, (b))
#endif

#ifndef OSF1
extern unsigned long inet_addr();
#endif
extern int socket(), connect();
extern int select();

static void dosubopt();
static void sendopt();

static int sending_size = 0;	/* true if we're communicating window size */

/*
 * connect to remote host
 * Warning: some voodoo code here
 */
int doconnect(addr, port)
char *addr; int port;
{
  struct sockaddr_in address;
  struct hostent *host_info;
  int err;
  
  bzero(&address, sizeof(address));
  /*
   * inet_addr has a strange design: It is documented to return -1 for
   * malformed requests, but it is declared to return unsigned long (or int) !
   * Anyway, this works.
   */    

#ifndef TERM  
  address.sin_addr.s_addr = inet_addr(addr);
  if(address.sin_addr.s_addr != (unsigned int)-1)
     address.sin_family = AF_INET;
  else
   {
     host_info = gethostbyname(addr);
     if(host_info == 0)
     	syserr("gethostbyname:unknown host");
     bcopy(host_info->h_addr, (char*)&address.sin_addr,
	   host_info->h_length);
     address.sin_family = host_info->h_addrtype;
   }
  address.sin_port = htons(port);
  
  sockfd = socket(address.sin_family, SOCK_STREAM, 0);
  if(sockfd < 0)
     syserr("creating socket");
  
  printf("#Trying %s... ", addr); fflush(stdout);
  while((err = connect(sockfd, (struct sockaddr *)&address,
		       sizeof(address))) < 0 && errno == EINTR);
#else /* term */
  
  if ((sockfd = connect_server(0)) < 0)
   {
     syserr("unable to connect to term server");
   }
  else
   {
     send_command(sockfd, C_PORT, 0, "%s:%d", addr, port);
     printf("Connected to term server...\n");
#ifdef TERM_COMPRESS
     send_command(sockfd, C_COMPRESS, 1, "y");
#endif     
     send_command(sockfd, C_DUMB, 1, 0);
   }
#endif /* term */
  
  if ((err < 0) || sockfd >= MAXFDSCAN) 
   {
     close(sockfd);
     return -1;
   }
  
  printf("connected!\n");
  return sockfd;
}

/*
 * read a maximum of maxbytes from remote host into buffer using the telnet
 * protocol. return bytes read.
 */
int read_socket(buffer, maxbytes)
char *buffer; int maxbytes;
{
    static enum {
	NORMAL, GOTIAC, GOTWILL, GOTWONT, GOTDO, GOTDONT, GOTSB, GOTSBIAC
    } state = NORMAL;
    static unsigned char buf[SOCKBUFSIZE];
    int i, n;
    unsigned char *p;
    static unsigned char subopt[MAXSUBOPT];
    int subchars;
    
    if(maxbytes > SOCKBUFSIZE) maxbytes = SOCKBUFSIZE;
    while((n = read(sockfd, buf, maxbytes)) < 0 && errno == EINTR);
    if(n == 0) {
	disconnect_sess(NULL);
	return 0;
    } else if(n < 0) {
	syserr("read from socket");
    }

    /* copy data to buffer and interpret telnet protocol escapes */

    for(i = 0, p = (unsigned char*)buffer; i < n; i++) {
	switch(state) {
	case NORMAL:
	    if(buf[i] == IAC)
		state = GOTIAC;
	    else
		*p++ = buf[i];
	    break;
	case GOTIAC:
	    switch(buf[i]) {
	    case WILL:
		state = GOTWILL; break;
	    case WONT:
		state = GOTWONT; break;
	    case DO:
		state = GOTDO; break;
	    case DONT:
		state = GOTDONT; break;
	    case SB:
		state = GOTSB;
		subchars = 0;
		break;
	    case IAC:
		*p++ = IAC;
		break;
	    case GA:
		/* I should handle GA as end-of-prompt marker one day */
	    default:
		/* ignore the rest of the telnet commands */
#ifdef TELOPTS
		printf("[skipped IAC <%d>]\n", buf[i]);
		*p++ = '@';
#endif
		state = NORMAL; break;
	    }
	    break;
	
	case GOTWILL:
#ifdef TELOPTS
	    printf("[got WILL %s]\n", TELOPTSTR(buf[i]));
#endif
	    switch(buf[i]) {
	    case TELOPT_ECHO:
		/* host echoes, turn off echo here */
		/*
		 * turn off only for main session, since we do not want
		 * subsidiary session password entries to block anything
		 * in the main session
		 */
		if (is_main_sess())
		    linemode |= LM_NOECHO;
		sendopt(DO, buf[i]);
		break;
	    case TELOPT_SGA:
		/* this can't hurt */
		linemode |= LM_CHAR;
		set_spec_keys();
		sendopt(DO, buf[i]);
		break;
	    default:
		/* don't accept other options */
		sendopt(DONT, buf[i]);
		break;
	    }
	    state = NORMAL;
	    break;
	
	case GOTWONT:
#ifdef TELOPTS
	    printf("[got WONT %s]\n", TELOPTSTR(buf[i]));
#endif
	    if(buf[i] == TELOPT_ECHO) {
		/* host no longer echoes, we do it instead */
		linemode &= ~LM_NOECHO;
	    }
	    /* accept any WONT */
	    sendopt(DONT, buf[i]);
	    state = NORMAL;
	    break;
	
	case GOTDO:
#ifdef TELOPTS
	    printf("[got DO %s]\n", TELOPTSTR(buf[i]));
#endif
	    switch(buf[i]) {
	    case TELOPT_SGA:
		linemode |= LM_CHAR;
		set_spec_keys();
		/* FALLTHROUGH */
	    case TELOPT_TTYPE:
		sendopt(WILL, buf[i]);
		break;
	    case TELOPT_NAWS:
		sendopt(WILL, buf[i]);
		sending_size = 1;
		send_term_size();
		break;
	    default:
		/* accept nothing else */
		sendopt(WONT, buf[i]);
		break;
	    }
	    state = NORMAL;
	    break;
	
	case GOTDONT:
#ifdef TELOPTS
	    printf("[got DONT %s]\n", TELOPTSTR(buf[i]));
#endif
	    if(buf[i] == TELOPT_SGA) {
		linemode &= ~LM_CHAR;
		set_spec_keys();
	    }
	    sendopt(WONT, buf[i]);
	    state = NORMAL;
	    break;
	
	case GOTSB:
	    if(buf[i] == IAC) {
		state = GOTSBIAC;
	    } else {
		if(subchars < MAXSUBOPT)
		    subopt[subchars++] = buf[i];
	    }
	    break;
	
	case GOTSBIAC:
	    if(buf[i] == IAC) {
		if(subchars < MAXSUBOPT)
		    subopt[subchars++] = IAC;
		state = GOTSB;
	    } else if(buf[i] == SE) {
		subopt[subchars] = '\0';
		dosubopt((char*)subopt);
		state = NORMAL;
	    } else {
		/* problem! I haven't the foggiest idea of what to do here.
		 * I'll just ignore it and hope it goes away. */
		printf("Telnet problem: got IAC <%d> instead of IAC SE!\n",
		       buf[i]);
		state = NORMAL;
	    }
	    break;
	default:
	    syserr("Canthappen!");
	    break;
	}
    }
    return (char*)p - buffer;
}

/*
 * send an option negotiation
 * 'what' is one of WILL, WONT, DO, DONT
 */
static void sendopt(what, opt)
int what; unsigned opt;
{
    unsigned char buf[3];
    buf[0] = IAC; buf[1] = what; buf[2] = opt;
    if(write(sockfd, buf, 3) < 3)
	syserr("write option");
#ifdef TELOPTS
    printf("[sent %s %s]\n", (what == WILL) ? "WILL" :
	   (what == WONT) ? "WONT" :
	   (what == DO) ? "DO" : (what == DONT) ? "DONT" : "error",
	   TELOPTSTR(opt));
#endif
}

/*
 * Send current terminal size (RFC 1073)
 */
void send_term_size()
{
    if(sending_size) {
	static unsigned char buf[9] =
	    { IAC, SB, TELOPT_NAWS, 0, 0, 0, 0, IAC, SE };

	buf[3] = cols >> 8;
	buf[4] = cols & 0xff;
	buf[5] = lines >> 8;
	buf[6] = lines & 0xff;
	if(write(main_sess(), buf, 9) < 9)
	    syserr("write term size");
#ifdef TELTOPS
	printf("[term size %d %d]\n", cols, lines);
#endif
    }
}

/*
 * process suboptions.
 * right now, only terminal type is processed.
 */
static void dosubopt(str)
char *str;
{
    char buf[256], *term;
  
    if(str[0] == TELOPT_TTYPE) {
	if(str[1] == 1) {
	    /* 1 == SEND */
#ifdef TELOPTS
	    printf("[got SB TERMINAL TYPE SEND]\n");
#endif
	    if(!(term = getenv("TERM"))) term = "unknown";
	    sprintf(buf, "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE, 0,
		    term, IAC, SE);	/* 0 == IS */
	    if(write(sockfd, buf, strlen(term) + 6) < strlen(term) + 6)
		syserr("write subopt to socket");
#ifdef TELOPTS
	    printf("[sent SB TERMINAL TYPE IS %s]\n", term);
#endif
	}
    }
}

static char output_buffer[OUTBUFSIZE];
static int output_len = 0;	/* number of characters in output_buffer */
static int output_socket = -1;	/* to which socket buffer should be sent*/

/*
 * put data in the output buffer for transmission to the remote host
 */
void send_to_host(data, len)
char *data; int len;
{
    int space;
    if(sockfd != output_socket) { /* is there data to another socket? */
	if(output_socket >= 0)
	    flush_output_buffer(); /* then flush it */
	output_socket = sockfd;
    }
    space = OUTBUFSIZE - output_len;
    if(space < len) {		/* not enough room? */
	bcopy(data, output_buffer + output_len, space);
	output_len = OUTBUFSIZE;
	flush_output_buffer();
	send_to_host(data + space, len - space);
    } else {
	bcopy(data, output_buffer + output_len, len);
	output_len += len;
    }
}

/*
 * send all buffered data to the remote host
 */
void flush_output_buffer()
{
    int n;
    char *p = output_buffer;
    while(output_len) {
	if((n = write(output_socket, p, output_len)) < 0)
	    syserr("write to socket");
	p += n;
	output_len -= n;
    }
}
